@@ -7,6 +7,7 @@ use crate::constant_pool::{ConstantPoolEntry, NameAndType, LiteralConstant, Meth
 use crate::constant_pool::{read_cp_utf8, read_cp_utf8_opt, read_cp_classinfo, read_cp_classinfo_opt, read_cp_nameandtype_opt,
     read_cp_literalconstant, read_cp_integer, read_cp_float, read_cp_long, read_cp_double, read_cp_methodhandle,
     read_cp_bootstrap_argument, read_cp_moduleinfo, read_cp_packageinfo};
+use crate::names::{is_field_descriptor, is_return_descriptor, is_unqualified_name};
 
 #[derive(Debug)]
 pub struct ExceptionTableEntry<'a> {
@@ -471,7 +472,13 @@ fn read_localvariable_data<'a>(bytes: &'a [u8], ix: &mut usize, pool: &[Rc<Const
         let start_pc = read_u2(bytes, ix)?;
         let length = read_u2(bytes, ix)?;
         let name = read_cp_utf8(bytes, ix, pool).map_err(|e| err!(e, "name for variable {}", i))?;
+        if !is_unqualified_name(&name, false, false) {
+            fail!("Invalid unqualified name for variable {}", i);
+        }
         let descriptor = read_cp_utf8(bytes, ix, pool).map_err(|e| err!(e, "descriptor for variable {}", i))?;
+        if !is_field_descriptor(&descriptor) {
+            fail!("Invalid descriptor for variable {}", i);
+        }
         let index = read_u2(bytes, ix)?;
         localvariables.push(LocalVariableEntry {
             start_pc,
@@ -491,6 +498,9 @@ fn read_localvariabletype_data<'a>(bytes: &'a [u8], ix: &mut usize, pool: &[Rc<C
         let start_pc = read_u2(bytes, ix)?;
         let length = read_u2(bytes, ix)?;
         let name = read_cp_utf8(bytes, ix, pool).map_err(|e| err!(e, "name for variable {}", i))?;
+        if !is_unqualified_name(&name, false, false) {
+            fail!("Invalid unqualified name for variable {}", i);
+        }
         let signature = read_cp_utf8(bytes, ix, pool).map_err(|e| err!(e, "signature for variable {}", i))?;
         let index = read_u2(bytes, ix)?;
         localvariabletypes.push(LocalVariableTypeEntry {
@@ -515,8 +525,21 @@ fn read_annotation_element_value<'a>(bytes: &'a [u8], ix: &mut usize, pool: &[Rc
         'S' => AnnotationElementValue::ShortConstant(read_cp_integer(bytes, ix, pool)?),
         'Z' => AnnotationElementValue::BooleanConstant(read_cp_integer(bytes, ix, pool)?),
         's' => AnnotationElementValue::StringConstant(read_cp_utf8(bytes, ix, pool)?),
-        'e' => AnnotationElementValue::EnumConstant { type_name: read_cp_utf8(bytes, ix, pool)?, const_name: read_cp_utf8(bytes, ix, pool)? },
-        'c' => AnnotationElementValue::ClassLiteral { class_name: read_cp_utf8(bytes, ix, pool)? },
+        'e' => {
+            let type_name = read_cp_utf8(bytes, ix, pool)?;
+            if !is_field_descriptor(&type_name) {
+                fail!("Invalid enum descriptor");
+            }
+            let const_name = read_cp_utf8(bytes, ix, pool)?;
+            AnnotationElementValue::EnumConstant { type_name, const_name }
+        }
+        'c' => {
+            let class_name = read_cp_utf8(bytes, ix, pool)?;
+            if !is_return_descriptor(&class_name) {
+                fail!("Invalid classinfo descriptor");
+            }
+            AnnotationElementValue::ClassLiteral { class_name }
+        }
         '@' => AnnotationElementValue::AnnotationValue(read_annotation(bytes, ix, pool)?),
         '[' => {
             let count = read_u2(bytes, ix)?;
@@ -533,6 +556,9 @@ fn read_annotation_element_value<'a>(bytes: &'a [u8], ix: &mut usize, pool: &[Rc
 
 fn read_annotation<'a>(bytes: &'a [u8], ix: &mut usize, pool: &[Rc<ConstantPoolEntry<'a>>]) -> Result<Annotation<'a>, ParseError> {
     let type_descriptor = read_cp_utf8(bytes, ix, pool).map_err(|e| err!(e, "type descriptor field"))?;
+    if !is_field_descriptor(&type_descriptor) {
+        fail!("Invalid descriptor");
+    }
     let element_count = read_u2(bytes, ix)?;
     let mut elements = Vec::with_capacity(element_count.into());
     for i in 0..element_count {
@@ -655,6 +681,9 @@ fn read_methodparameters_data<'a>(bytes: &'a [u8], ix: &mut usize, pool: &[Rc<Co
     let mut methodparameters = Vec::with_capacity(count.into());
     for i in 0..count {
         let name = read_cp_utf8_opt(bytes, ix, pool).map_err(|e| err!(e, "name of method parameter {}", i))?;
+        if name.is_some() && !is_unqualified_name(name.as_ref().unwrap(), false, false) {
+            fail!("Invalid unqualified name for variable {}", i);
+        }
         let access_flags = MethodParameterAccessFlags::from_bits(read_u2(bytes, ix)?).ok_or_else(|| err!(("Invalid access flags found"), ("method parameter {}", i)))?;
         methodparameters.push(MethodParameterEntry {
             name,
@@ -763,7 +792,13 @@ fn read_record_data<'a>(bytes: &'a [u8], ix: &mut usize, pool: &[Rc<ConstantPool
     let mut components = Vec::with_capacity(count.into());
     for i in 0..count {
         let name = read_cp_utf8(bytes, ix, pool).map_err(|e| err!(e, "name of entry {}", i))?;
+        if !is_unqualified_name(&name, false, false) {
+            fail!("Invalid unqualified name for entry {}", i);
+        }
         let descriptor = read_cp_utf8(bytes, ix, pool).map_err(|e| err!(e, "descriptor of entry {}", i))?;
+        if !is_field_descriptor(&descriptor) {
+            fail!("Invalid descriptor for entry {}", i);
+        }
         let attributes = read_attributes(bytes, ix, pool).map_err(|e| err!(e, "entry {}", i))?;
         components.push(RecordComponentEntry {
             name,
@@ -817,6 +852,7 @@ pub(crate) fn read_attributes<'a>(bytes: &'a [u8], ix: &mut usize, pool: &[Rc<Co
             }
             "Signature" => {
                 ensure_length(length, 2).map_err(|e| err!(e, "Signature attribute {}", i))?;
+                // TODO: validate signature
                 AttributeData::Signature(read_cp_utf8(bytes, ix, pool).map_err(|e| err!(e, "signature field of Signature attribute {}", i))?)
             }
             "SourceFile" => {
