@@ -4,6 +4,7 @@ use std::ops::Deref;
 use std::rc::Rc;
 
 use crate::{read_u1, read_u2, read_u4, read_u8, ParseError};
+use crate::names::{is_array_descriptor, is_binary_name, is_field_descriptor, is_method_descriptor, is_module_name, is_unqualified_name};
 
 #[derive(Debug)]
 pub(crate) enum ConstantPoolRef<'a> {
@@ -193,12 +194,29 @@ impl<'a> ConstantPoolEntry<'a> {
 
     fn validate(&self, major_version: u16) -> Result<bool, ParseError> {
         match self {
-            ConstantPoolEntry::ClassInfo(x) => x.ensure_type(ConstantPoolEntryTypes::UTF8),
+            ConstantPoolEntry::ClassInfo(x) => Ok(x.ensure_type(ConstantPoolEntryTypes::UTF8)? && x.borrow().get().validate_classinfo_name()?),
             ConstantPoolEntry::String(x) => x.ensure_type(ConstantPoolEntryTypes::UTF8),
-            ConstantPoolEntry::FieldRef(x, y) => Ok(x.ensure_type(ConstantPoolEntryTypes::CLASS_INFO)? && y.ensure_type(ConstantPoolEntryTypes::NAME_AND_TYPE)?),
-            ConstantPoolEntry::MethodRef(x, y) => Ok(x.ensure_type(ConstantPoolEntryTypes::CLASS_INFO)? && y.ensure_type(ConstantPoolEntryTypes::NAME_AND_TYPE)?),
-            ConstantPoolEntry::InterfaceMethodRef(x, y) => Ok(x.ensure_type(ConstantPoolEntryTypes::CLASS_INFO)? && y.ensure_type(ConstantPoolEntryTypes::NAME_AND_TYPE)?),
-            ConstantPoolEntry::NameAndType(x, y) => Ok(x.ensure_type(ConstantPoolEntryTypes::UTF8)? && y.ensure_type(ConstantPoolEntryTypes::UTF8)?),
+            ConstantPoolEntry::FieldRef(x, y) => Ok(
+                x.ensure_type(ConstantPoolEntryTypes::CLASS_INFO)? &&
+                y.ensure_type(ConstantPoolEntryTypes::NAME_AND_TYPE)? &&
+                y.borrow().get().validate_field_descriptor()?
+            ),
+            ConstantPoolEntry::MethodRef(x, y) => Ok(
+                x.ensure_type(ConstantPoolEntryTypes::CLASS_INFO)? &&
+                y.ensure_type(ConstantPoolEntryTypes::NAME_AND_TYPE)? &&
+                y.borrow().get().validate_method_descriptor()?
+            ),
+            ConstantPoolEntry::InterfaceMethodRef(x, y) => Ok(
+                x.ensure_type(ConstantPoolEntryTypes::CLASS_INFO)? &&
+                y.ensure_type(ConstantPoolEntryTypes::NAME_AND_TYPE)? &&
+                y.borrow().get().validate_method_descriptor()?
+            ),
+            ConstantPoolEntry::NameAndType(x, y) => Ok(
+                x.ensure_type(ConstantPoolEntryTypes::UTF8)? &&
+                x.borrow().get().validate_unqualified_name()? &&
+                y.ensure_type(ConstantPoolEntryTypes::UTF8)?
+                // y is validated as part of FieldRef/MethodRef/InterfaceMethodRef/Dynamic/InvokeDynamic pool item validation
+            ),
             ConstantPoolEntry::MethodHandle(x, y) => y.ensure_type(match x {
                 ReferenceKind::GetField |
                 ReferenceKind::GetStatic |
@@ -210,11 +228,17 @@ impl<'a> ConstantPoolEntry<'a> {
                 ReferenceKind::InvokeSpecial => if major_version < 52 { ConstantPoolEntryTypes::METHOD_REF } else { ConstantPoolEntryTypes::NEW_METHOD_REFS },
                 ReferenceKind::InvokeInterface => ConstantPoolEntryTypes::INTERFACE_METHOD_REF,
             }),
-            ConstantPoolEntry::MethodType(x) => x.ensure_type(ConstantPoolEntryTypes::UTF8),
-            ConstantPoolEntry::Dynamic(_, y) => y.ensure_type(ConstantPoolEntryTypes::NAME_AND_TYPE),
-            ConstantPoolEntry::InvokeDynamic(_, y) => y.ensure_type(ConstantPoolEntryTypes::NAME_AND_TYPE),
-            ConstantPoolEntry::ModuleInfo(x) => x.ensure_type(ConstantPoolEntryTypes::UTF8),
-            ConstantPoolEntry::PackageInfo(x) => x.ensure_type(ConstantPoolEntryTypes::UTF8),
+            ConstantPoolEntry::MethodType(x) => Ok(x.ensure_type(ConstantPoolEntryTypes::UTF8)? && x.borrow().get().validate_method_descriptor()?),
+            ConstantPoolEntry::Dynamic(_, y) => Ok(
+                y.ensure_type(ConstantPoolEntryTypes::NAME_AND_TYPE)? &&
+                y.borrow().get().validate_field_descriptor()?
+            ),
+            ConstantPoolEntry::InvokeDynamic(_, y) => Ok(
+                y.ensure_type(ConstantPoolEntryTypes::NAME_AND_TYPE)? &&
+                y.borrow().get().validate_method_descriptor()?
+            ),
+            ConstantPoolEntry::ModuleInfo(x) => Ok(x.ensure_type(ConstantPoolEntryTypes::UTF8)? && x.borrow().get().validate_module_name()?),
+            ConstantPoolEntry::PackageInfo(x) => Ok(x.ensure_type(ConstantPoolEntryTypes::UTF8)? && x.borrow().get().validate_binary_name()?),
             _ => Ok(true),
         }
     }
@@ -227,6 +251,98 @@ impl<'a> ConstantPoolEntry<'a> {
         }
     }
 
+    fn validate_classinfo_name(&self) -> Result<bool, ParseError> {
+        match self {
+            ConstantPoolEntry::Utf8(x) => {
+                // Per 4.4.1, classinfo names are allowed to be array descriptors too. This happens in the java 16 modules file.
+                if is_binary_name(x) || is_array_descriptor(x) {
+                    Ok(true)
+                } else {
+                    fail!("Invalid binary name")
+                }
+            }
+            _ => panic!("Attempting to get utf-8 data from non-utf8 constant pool entry!"),
+        }
+    }
+
+    fn validate_binary_name(&self) -> Result<bool, ParseError> {
+        match self {
+            ConstantPoolEntry::Utf8(x) => {
+                if is_binary_name(x) {
+                    Ok(true)
+                } else {
+                    fail!("Invalid binary name")
+                }
+            }
+            _ => panic!("Attempting to get utf-8 data from non-utf8 constant pool entry!"),
+        }
+    }
+
+    fn validate_unqualified_name(&self) -> Result<bool, ParseError> {
+        match self {
+            ConstantPoolEntry::Utf8(x) => {
+                if is_unqualified_name(x, true, false) {
+                    Ok(true)
+                } else {
+                    fail!("Invalid unqualified name")
+                }
+            }
+            _ => panic!("Attempting to get utf-8 data from non-utf8 constant pool entry!"),
+        }
+    }
+
+    fn validate_module_name(&self) -> Result<bool, ParseError> {
+        match self {
+            ConstantPoolEntry::Utf8(x) => {
+                if is_module_name(x) {
+                    Ok(true)
+                } else {
+                    fail!("Invalid module name")
+                }
+            }
+            _ => panic!("Attempting to get utf-8 data from non-utf8 constant pool entry!"),
+        }
+    }
+
+    fn validate_field_descriptor(&self) -> Result<bool, ParseError> {
+        match self {
+            ConstantPoolEntry::NameAndType(_, y) => {
+                // Call ensure_type to fail with an error rather than panicking if we happen to be
+                // in the process of validating a FieldRef constant pool entry whose NameAndType
+                // points to a later entry in the constant pool that hasn't been validated yet.
+                // assert on the bool because we should never get Ok(false).
+                assert!(y.ensure_type(ConstantPoolEntryTypes::UTF8)?);
+                if is_field_descriptor(&y.borrow().get().utf8()) {
+                    Ok(true)
+                } else {
+                    fail!("Invalid field descriptor")
+                }
+            }
+            _ => panic!("Attempting to get descriptor from non-NameAndType constant pool entry!"),
+        }
+    }
+
+    fn validate_method_descriptor(&self) -> Result<bool, ParseError> {
+        match self {
+            ConstantPoolEntry::NameAndType(_, y) => {
+                // Call ensure_type to fail with an error rather than panicking if we happen to be
+                // in the process of validating a [Interface]MethodRef constant pool entry whose NameAndType
+                // points to a later entry in the constant pool that hasn't been validated yet.
+                // assert on the bool because we should never get Ok(false).
+                assert!(y.ensure_type(ConstantPoolEntryTypes::UTF8)?);
+                y.borrow().get().validate_method_descriptor()
+            }
+            ConstantPoolEntry::Utf8(x) => {
+                if is_method_descriptor(&x) {
+                    Ok(true)
+                } else {
+                    fail!("Invalid method descriptor")
+                }
+            }
+            _ => panic!("Attempting to get descriptor from non-NameAndType constant pool entry!"),
+        }
+    }
+
     fn utf8(&self) -> Cow<'a, str> {
         match self {
             ConstantPoolEntry::Utf8(x) => x.clone(),
@@ -396,7 +512,8 @@ fn resolve_constant_pool<'a>(constant_pool: &[Rc<ConstantPoolEntry<'a>>]) -> Res
 
 fn validate_constant_pool<'a>(constant_pool: &[Rc<ConstantPoolEntry<'a>>], major_version: u16) -> Result<(), ParseError> {
     for (i, cp_entry) in constant_pool.iter().enumerate() {
-        cp_entry.validate(major_version).map_err(|e| err!(e, "constant pool entry {}", i))?;
+        let valid = cp_entry.validate(major_version).map_err(|e| err!(e, "constant pool entry {}", i))?;
+        assert!(valid); // validate functions should never return Ok(false)
     }
     Ok(())
 }
